@@ -17,7 +17,7 @@ import (
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
 	"k8s.io/apimachinery/pkg/runtime"
-	"k8s.io/apimachinery/pkg/runtime/schema"
+	"k8s.io/apimachinery/pkg/types"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/controller"
 	"sigs.k8s.io/controller-runtime/pkg/handler"
@@ -78,17 +78,15 @@ func add(mgr manager.Manager, r reconcile.Reconciler) error {
 	}
 
 	// Watch for changes to primary resource KnativeServing
-	err = c.Watch(&source.Kind{Type: &servingv1alpha1.KnativeServing{}}, &handler.EnqueueRequestForObject{}, predicate.GenerationChangedPredicate{})
-	if err != nil {
+	if err = c.Watch(&source.Kind{Type: &servingv1alpha1.KnativeServing{}}, &handler.EnqueueRequestForObject{}, predicate.GenerationChangedPredicate{}); err != nil {
 		return err
 	}
 
 	// Watch child deployments for availability
-	err = c.Watch(&source.Kind{Type: &appsv1.Deployment{}}, &handler.EnqueueRequestForOwner{
+	if err = c.Watch(&source.Kind{Type: &appsv1.Deployment{}}, &handler.EnqueueRequestForOwner{
 		IsController: true,
 		OwnerType:    &servingv1alpha1.KnativeServing{},
-	})
-	if err != nil {
+	}); err != nil {
 		return err
 	}
 
@@ -137,9 +135,6 @@ func (r *ReconcileKnativeServing) Reconcile(request reconcile.Request) (reconcil
 	instance := &servingv1alpha1.KnativeServing{}
 	if err := r.client.Get(context.TODO(), request.NamespacedName, instance); err != nil {
 		if apierrors.IsNotFound(err) {
-			if isInteresting(request) {
-				r.config.DeleteAll()
-			}
 			return reconcile.Result{}, nil
 		}
 		return reconcile.Result{}, err
@@ -149,9 +144,13 @@ func (r *ReconcileKnativeServing) Reconcile(request reconcile.Request) (reconcil
 		return reconcile.Result{}, r.ignore(instance)
 	}
 
+	if instance.GetDeletionTimestamp() != nil {
+		return reconcile.Result{}, r.delete(instance)
+	}
+
 	stages := []func(*servingv1alpha1.KnativeServing) error{
+		r.ensureFinalizers,
 		r.initStatus,
-		r.checkDependencies,
 		r.install,
 		r.checkDeployments,
 		r.deleteObsoleteResources,
@@ -159,6 +158,9 @@ func (r *ReconcileKnativeServing) Reconcile(request reconcile.Request) (reconcil
 
 	for _, stage := range stages {
 		if err := stage(instance); err != nil {
+			if _, ok := err.(*common.NotYetReadyError); ok {
+				return reconcile.Result{}, nil
+			}
 			return reconcile.Result{}, err
 		}
 	}
@@ -190,6 +192,17 @@ func (r *ReconcileKnativeServing) updateStatus(instance *servingv1alpha1.Knative
 	return nil
 }
 
+func (r *ReconcileKnativeServing) ensureFinalizers(instance *servingv1alpha1.KnativeServing) error {
+	for _, finalizer := range instance.GetFinalizers() {
+		if finalizer == operand {
+			return nil
+		}
+	}
+
+	instance.SetFinalizers(append(instance.GetFinalizers(), operand))
+	return r.client.Update(context.TODO(), instance)
+}
+
 // Apply the embedded resources
 func (r *ReconcileKnativeServing) install(instance *servingv1alpha1.KnativeServing) error {
 	defer r.updateStatus(instance)
@@ -205,7 +218,11 @@ func (r *ReconcileKnativeServing) install(instance *servingv1alpha1.KnativeServi
 		}
 	}
 	if err != nil {
-		instance.Status.MarkInstallFailed(err.Error())
+		if _, ok := err.(*common.NotYetReadyError); ok {
+			instance.Status.MarkInstallNotReady("Install in progress: " + err.Error())
+			return err
+		}
+		instance.Status.MarkInstallFailed("Install failed with message: " + err.Error())
 		return err
 	}
 
@@ -216,6 +233,26 @@ func (r *ReconcileKnativeServing) install(instance *servingv1alpha1.KnativeServi
 	return nil
 }
 
+func (r *ReconcileKnativeServing) delete(instance *servingv1alpha1.KnativeServing) error {
+	if len(instance.GetFinalizers()) == 0 || instance.GetFinalizers()[0] != operand {
+		return nil
+	}
+	if err := r.extensions.Finalize(instance); err != nil {
+		return err
+	}
+	if err := r.config.DeleteAll(); err != nil {
+		return err
+	}
+
+	// The deletionTimestamp might've changed. Fetch the resource again.
+	refetched := &servingv1alpha1.KnativeServing{}
+	if err := r.client.Get(context.TODO(), types.NamespacedName{Namespace: instance.Namespace, Name: instance.Name}, refetched); err != nil {
+		return err
+	}
+	refetched.SetFinalizers(refetched.GetFinalizers()[1:])
+	return r.client.Update(context.TODO(), refetched)
+}
+
 // Expose metrics for installed knative serving operator
 func (r *ReconcileKnativeServing) exposeMetrics(instance *servingv1alpha1.KnativeServing) {
 	if instance.Status.GetConditions() != nil {
@@ -262,25 +299,6 @@ func (r *ReconcileKnativeServing) checkDeployments(instance *servingv1alpha1.Kna
 	return nil
 }
 
-// Check for all dependencies
-func (r *ReconcileKnativeServing) checkDependencies(instance *servingv1alpha1.KnativeServing) error {
-	defer r.updateStatus(instance)
-
-	istio := schema.GroupVersionKind{Group: "networking.istio.io", Version: "v1alpha3", Kind: "gateway"}
-	list := &unstructured.UnstructuredList{}
-	list.SetGroupVersionKind(istio)
-	if err := r.client.List(context.TODO(), nil, list); err != nil {
-		msg := fmt.Sprintf("Istio not detected, GVK %v missing", istio)
-		instance.Status.MarkDependencyMissing(msg)
-		log.Error(err, msg)
-		return err
-	}
-
-	log.Info("All dependencies are installed")
-	instance.Status.MarkDependenciesInstalled()
-	return nil
-}
-
 // Delete obsolete resources from previous versions
 func (r *ReconcileKnativeServing) deleteObsoleteResources(instance *servingv1alpha1.KnativeServing) error {
 	// istio-system resources from 0.3
@@ -341,7 +359,7 @@ func isInteresting(request reconcile.Request) bool {
 func (r *ReconcileKnativeServing) ignore(instance *servingv1alpha1.KnativeServing) (err error) {
 	err = r.initStatus(instance)
 	if err == nil {
-		msg := fmt.Sprintf("The only KnativeServing resource that matters is %s/%s", operand, operand)
+		msg := fmt.Sprintf("The KnativeServing resource needs to be created as %s/%s, otherwise it will be ignored", operand, operand)
 		instance.Status.MarkIgnored(msg)
 		err = r.updateStatus(instance)
 	}
